""" Command line options for Cumulus DS """
import argparse


# Read arguments from the command line
PARSER = argparse.ArgumentParser(
    description='Cumulus cloud management tool')
GENERAL_AG = PARSER.add_argument_group('General options')
GENERAL_AG.add_argument(
    '-e', '--environment',
    help='Environment to use')
GENERAL_AG.add_argument(
    '-s', '--stacks',
    help=(
        'Comma separated list of stacks to deploy. '
        'Default behavior is to deploy all stacks for an environment'))
GENERAL_AG.add_argument(
    '--version',
    help=(
        'Environment version number. '
        'Overrides the version value from the configuration file'))
GENERAL_AG.add_argument(
    '--parameters',
    help=(
        'CloudFormation parameters. On the form: '
        'stack_name:parameter_name=value,stack_name=parameter_name=value'
    ))
GENERAL_AG.add_argument(
    '--config',
    help='Path to configuration file.')
GENERAL_AG.add_argument(
    '--cumulus-version',
    action='count',
    help='Print cumulus version number')
GENERAL_AG.add_argument(
    '--force',
    default=False,
    action='store_true',
    help='Skip any safety questions')
ACTIONS_AG = PARSER.add_argument_group('Actions')
ACTIONS_AG.add_argument(
    '--bundle',
    action='count',
    help='Build and upload bundles to AWS S3')
ACTIONS_AG.add_argument(
    '--deploy',
    action='count',
    help='Bundle and deploy all stacks in the environment')
ACTIONS_AG.add_argument(
    '--deploy-without-bundling',
    action='count',
    help='Deploy all stacks in the environment, without bundling first')
ACTIONS_AG.add_argument(
    '--events',
    action='count',
    help='List events for the stack')
ACTIONS_AG.add_argument(
    '--list',
    action='count',
    help='List stacks for each environment')
ACTIONS_AG.add_argument(
    '--outputs',
    action='count',
    help='Show output for all stacks')
ACTIONS_AG.add_argument(
    '--validate-templates',
    action='count',
    help='Validate all templates for the environment')
ACTIONS_AG.add_argument(
    '--undeploy',
    action='count',
    help=(
        'Undeploy (delete) all stacks in the environment. '
        'Use --force to skip the safety question.'))



""" Connection handler """
import boto
import logging
from boto import cloudformation

from cumulus_ds.config import CONFIG as config

logger = logging.getLogger(__name__)


def connect_s3():
    """ Connect to AWS S3

    :returns: boto.s3.connection
    """
    try:
        return boto.connect_s3(
            aws_access_key_id=config.get_environment_option(
                'access-key-id'),
            aws_secret_access_key=config.get_environment_option(
                'secret-access-key'))
    except Exception as err:
        logger.error('A problem occurred connecting to AWS S3: {}'.format(err))
        raise


def connect_cloudformation():
    """ Connect to AWS CloudFormation

    :returns: boto.cloudformation.connection
    """
    try:
        return cloudformation.connect_to_region(
            config.get_environment_option('region'),
            aws_access_key_id=config.get_environment_option(
                'access-key-id'),
            aws_secret_access_key=config.get_environment_option(
                'secret-access-key'))
    except Exception as err:
        logger.error(
            'A problem occurred connecting to AWS CloudFormation: {}'.format(
                err))
        raise



""" Functions for calculating the teminal size """
import os
import shlex
import struct
import platform
import subprocess


def get_terminal_size():
    """ Get the current terminal size """
    current_os = platform.system()
    tuple_xy = None
    if current_os == 'Windows':
        tuple_xy = _get_terminal_size_windows()
        if tuple_xy is None:
            tuple_xy = _get_terminal_size_tput()
            # needed for window's python in cygwin's xterm!
    if current_os in ['Linux', 'Darwin'] or current_os.startswith('CYGWIN'):
        tuple_xy = _get_terminal_size_linux()
    if tuple_xy is None:
        print "default"
        tuple_xy = (80, 25)      # default value
    return tuple_xy


def _get_terminal_size_windows():
    """ Get the terminal size on Windows """
    try:
        from ctypes import windll, create_string_buffer
        # stdin handle is -10
        # stdout handle is -11
        # stderr handle is -12
        handle = windll.kernel32.GetStdHandle(-12)
        csbi = create_string_buffer(22)
        res = windll.kernel32.GetConsoleScreenBufferInfo(handle, csbi)
        if res:
            (_, _, _, _, _,
             left, top, right, bottom,
             _, _) = struct.unpack("hhhhHhhhhhh", csbi.raw)
            sizex = right - left + 1
            sizey = bottom - top + 1
            return sizex, sizey
    except:
        pass


def _get_terminal_size_tput():
    """ Fallback function for Windows """
    try:
        cols = int(subprocess.check_call(shlex.split('tput cols')))
        rows = int(subprocess.check_call(shlex.split('tput lines')))
        return (cols, rows)
    except:
        pass


def _get_terminal_size_linux():
    """ Get the terminal size in Linux / Darwin """
    def ioctl_GWINSZ(fdd):
        """ Undocumented """
        try:
            import fcntl
            import termios
            crd = struct.unpack(
                'hh',
                fcntl.ioctl(fdd, termios.TIOCGWINSZ, '1234'))
            return crd
        except:
            pass
    crd = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
    if not crd:
        try:
            fdd = os.open(os.ctermid(), os.O_RDONLY)
            crd = ioctl_GWINSZ(fdd)
            os.close(fdd)
        except:
            pass
    if not crd:
        try:
            crd = (os.environ['LINES'], os.environ['COLUMNS'])
        except:
            return None
    return int(crd[1]), int(crd[0])



class ChecksumMismatchException(Exception):
    """ A checksum check has failed """
    pass


class ConfigurationException(Exception):
    """ Configuration exception """
    pass


class HookExecutionException(Exception):
    """ Failed to execute a hook """
    pass


class InvalidTemplateException(Exception):
    """ Invalid CloudFormation template """
    pass


class UnsupportedCompression(Exception):
    """ An unsupported compression format for the bundle found """
    pass



""" Stack helpers """
import json
import logging
import sys
import time
from datetime import datetime, timedelta

import boto

from cumulus_ds import connection_handler
from cumulus_ds import terminal_size
from cumulus_ds.config import CONFIG as config
from cumulus_ds.exceptions import InvalidTemplateException

if sys.platform in ['win32', 'cygwin']:
    import ntpath as ospath
else:
    import os.path as ospath

LOGGER = logging.getLogger(__name__)
CONNECTION = connection_handler.connect_cloudformation()
TERMINAL_WIDTH, _ = terminal_size.get_terminal_size()


def delete_stack(stack):
    """ Delete an existing stack

    :type stack: str
    :param stack: Stack name
    """
    LOGGER.info('Deleting stack {}'.format(stack))
    CONNECTION.delete_stack(stack)
    _wait_for_stack_complete(stack, filter_type='DELETE')


def ensure_stack(
        stack_name, parameters, template, tags=None, disable_rollback=False,
        timeout_in_minutes=None, capabilities=['CAPABILITY_IAM']):
    """ Ensure that a CloudFormation stack is running

    If the stack does not exist, it will be created. If the stack exists
    it will be updated.

    :type stack_name: str
    :param stack_name: Name of the stack to ensure
    :type parameters: list
    :param parameters: List of tuples with parameters and values
    :type template: str
    :parameter template: Template in JSON string or a HTTP URL
    :type tags: dict
    :param tags: Dict with keys and values
    :type disable_rollback: bool
    :param disable_rollback: Disable rollbacks of failed creates/updates
    :type timeout_in_minutes: int
    :parameter timeout_in_minutes: Timeout the stack creation after x minutes
    :type capabilities: list
    :parameter capabilities: The list of capabilities you want to allow in the
        stack. Currently, the only valid capability is 'CAPABILITY_IAM'
    """
    LOGGER.info('Ensuring stack {} with template {}'.format(
        stack_name, template))

    cumulus_parameters = [
        ('CumulusBundleBucket', config.get_environment_option('bucket')),
        ('CumulusEnvironment', config.get_environment()),
        ('CumulusVersion', config.get_environment_option('version'))
    ]

    for parameter in cumulus_parameters + parameters:
        LOGGER.debug(
            'Adding parameter "{}" with value "{}" to CF template'.format(
                parameter[0], parameter[1]))

    if timeout_in_minutes:
        LOGGER.debug('Will time out stack creation after {:d} minutes'.format(
            timeout_in_minutes))

    try:
        if stack_exists(stack_name):
            LOGGER.debug('Updating existing stack to version {}'.format(
                config.get_environment_option('version')))

            if template[0:4] == 'http':
                CONNECTION.update_stack(
                    stack_name,
                    parameters=cumulus_parameters + parameters,
                    template_url=template,
                    disable_rollback=disable_rollback,
                    capabilities=['CAPABILITY_IAM'],
                    timeout_in_minutes=timeout_in_minutes,
                    tags=tags)
            else:
                CONNECTION.update_stack(
                    stack_name,
                    parameters=cumulus_parameters + parameters,
                    template_body=_get_json_from_template(template),
                    disable_rollback=disable_rollback,
                    capabilities=['CAPABILITY_IAM'],
                    timeout_in_minutes=timeout_in_minutes,
                    tags=tags)

            _wait_for_stack_complete(stack_name, filter_type='UPDATE')
        else:
            LOGGER.debug('Creating new stack with version {}'.format(
                config.get_environment_option('version')))
            if template[0:4] == 'http':
                CONNECTION.create_stack(
                    stack_name,
                    parameters=cumulus_parameters + parameters,
                    template_url=template,
                    disable_rollback=disable_rollback,
                    capabilities=['CAPABILITY_IAM'],
                    timeout_in_minutes=timeout_in_minutes,
                    tags=tags)
            else:
                CONNECTION.create_stack(
                    stack_name,
                    parameters=cumulus_parameters + parameters,
                    template_body=_get_json_from_template(template),
                    disable_rollback=disable_rollback,
                    capabilities=['CAPABILITY_IAM'],
                    timeout_in_minutes=timeout_in_minutes,
                    tags=tags)

            _wait_for_stack_complete(stack_name, filter_type='CREATE')

    except ValueError as error:
        raise InvalidTemplateException(
            'Malformatted template: {}'.format(error))
    except boto.exception.BotoServerError as error:
        code = eval(error.error_message)['Error']['Code']
        message = eval(error.error_message)['Error']['Message']

        if code == 'ValidationError':
            if message == 'No updates are to be performed.':
                # Do not raise this exception if it is due to lack of updates
                # We do not want to fail any other stack updates after this
                # stack
                LOGGER.warning(
                    'No CloudFormation updates are to be '
                    'performed for {}'.format(stack_name))
                return

        LOGGER.error('Boto exception {}: {}'.format(code, message))
        return

    _print_stack_output(stack_name)


def get_stack_by_name(stack_name):
    """ Returns a stack given its name

    :type stack_name: str
    :param stack_name: Stack name
    :returns: stack or None
    """
    for stack in CONNECTION.list_stacks():
        if (stack.stack_status != 'DELETE_COMPLETE' and
                stack.stack_name == stack_name):
            return stack

    return None


def list_events_all_stacks():
    """ List events for all configured stacks """
    for stack_name in config.get_stacks():
        stack = get_stack_by_name(stack_name)

        if not stack:
            break

        _print_event_log_title()

        written_events = []
        for event in reversed(CONNECTION.describe_stack_events(stack.stack_id)):
            if event.event_id not in written_events:
                written_events.append(event.event_id)
                _print_event_log_event(event)


def list_all_stacks():
    """ List stacks and their statuses """
    cf_stacks = CONNECTION.list_stacks()

    for stack in config.get_stacks():
        stack_found = False
        for cf_stack in cf_stacks:
            if stack == cf_stack.stack_name:
                stack_found = True

        if stack_found:
            print('{:<30}{}'.format(stack, cf_stack.stack_status))
        else:
            print('{:<30}{}'.format(stack, 'NOT_RUNNING'))


def print_output_all_stacks():
    """ Print the output for all stacks """
    for stack in config.get_stacks():
        _print_stack_output(stack)


def stack_exists(stack_name):
    """ Check if a stack exists

    :type stack_name: str
    :param stack_name: Stack name
    :returns: bool
    """
    for stack in CONNECTION.list_stacks():
        if (stack.stack_status != 'DELETE_COMPLETE' and
                stack.stack_name == stack_name):
            return True

    return False


def validate_templates_all_stacks():
    """ Validate the template for all stacks"""
    for stack in config.get_stacks():
        template = config.get_stack_template(stack)

        result = CONNECTION.validate_template(_get_json_from_template(template))
        if result:
            LOGGER.info('Template {} is valid!'.format(template))


def _get_stack_outputs(stack_name_or_id):
    """ Get a list of stack output values

    :type stack_name_or_id: str
    :param stack_name_or_id: Stack name or id
    :returns: list -- List of (key, value)
    """
    try:
        stack = CONNECTION.describe_stacks(stack_name_or_id)[0]
    except IndexError:
        LOGGER.debug(
            'No stack with name or id {} found'.format(stack_name_or_id))
        return []

    return stack.outputs


def _get_json_from_template(template):
    """ Returns a JSON string given a template file path

    :type template: str
    :param template: Template path to use
    :returns: JSON object
    """
    template_path = ospath.expandvars(ospath.expanduser(template))
    LOGGER.debug('Parsing template file {}'.format(template_path))

    file_handle = open(template_path)
    json_data = json.dumps(json.loads(file_handle.read()))
    file_handle.close()

    return json_data


def _print_event_log_event(event):
    """ Print event log row to stdout

    :type event: event object
    :param event: CloudFormation event object
    """
    # Colorize status
    event_status = event.resource_status.split('_')
    if event_status[len(event_status) - 1] == 'COMPLETE':
        # Green text
        status = '\033[92m' + event.resource_status + '\033[0m'
    elif event_status[len(event_status) - 1] == 'PROGRESS':
        # Blue text
        status = '\033[94m' + event.resource_status + '\033[0m'
    elif event_status[len(event_status) - 1] == 'FAILED':
        # Red text
        status = '\033[91m' + event.resource_status + '\033[0m'
    else:
        status = event.resource_status

    row = '{timestamp:<19}'.format(
        timestamp=datetime.strftime(event.timestamp, '%Y-%m-%d %H:%M:%S'))
    row += ' | {type:<45}'.format(type=event.resource_type)
    row += ' | {logical_id:<42}'.format(logical_id=event.logical_resource_id)

    if TERMINAL_WIDTH >= 190:
        if event.resource_status_reason:
            reason = event.resource_status_reason
        else:
            reason = ''

        row += ' | {reason:<36}'.format(reason=reason)

    row += ' | {status:<33}'.format(status=status.replace('_', ' ').lower())

    print(row)


def _print_event_log_separator():
    """ Print separator line for the event log """
    row = '--------------------'  # Timestamp
    row += '+-----------------------------------------------'  # Resource type
    row += '+--------------------------------------------'  # Logical ID

    if TERMINAL_WIDTH >= 190:
        row += '+--------------------------------------'  # Reason

    row += '+--------------------------------'  # Status

    print(row)


def _print_event_log_title():
    """ Print event log title row on stdout """
    _print_event_log_separator()

    row = '{timestamp:<19}'.format(timestamp='Timestamp')
    row += ' | {type:<45}'.format(type='Resource type')
    row += ' | {logical_id:<42}'.format(logical_id='Logical ID')
    if TERMINAL_WIDTH >= 190:
        row += ' | {reason:<36}'.format(reason='Reason')
    row += ' | {status:<25}'.format(status='Status')

    print(row)

    _print_event_log_separator()


def _print_stack_output(stack_name_or_id):
    """ Print the stack output for a given stack

    :type stack_name_or_id: str
    :param stack_name_or_id: Stack name
    :returns: None
    """
    LOGGER.info('Printing output for stack "{}"'.format(stack_name_or_id))
    outputs = _get_stack_outputs(stack_name_or_id)

    if not outputs:
        LOGGER.info('No outputs found for stack "{}"'.format(stack_name_or_id))
        return

    print(
        '---------------------+----------------------------------'
        '--------------------------------------------------------')
    print('{key_title:<20} | {value_title:<45}'.format(
        key_title='Tag',
        value_title='Value'))
    print(
        '---------------------+----------------------------------'
        '--------------------------------------------------------')
    for output in outputs:
        print('{key:<20} | {value:<45}'.format(
            key=output.key,
            value=output.value))
    print(
        '---------------------+----------------------------------'
        '--------------------------------------------------------')


def _wait_for_stack_complete(stack_name, check_interval=5, filter_type=None):
    """ Wait until the stack create/update has been completed

    :type stack_name: str
    :param stack_name: Stack name
    :type check_interval: int
    :param check_interval: Seconds between each console update
    :type filter_type: str
    :param filter_type: Filter events by type. Supported values are None,
        CREATE, DELETE, UPDATE. Rollback events are always shown.
    """
    start_time = datetime.utcnow() - timedelta(0, 10)
    complete = False
    complete_statuses = [
        'CREATE_FAILED',
        'CREATE_COMPLETE',
        'ROLLBACK_FAILED',
        'ROLLBACK_COMPLETE',
        'DELETE_FAILED',
        'DELETE_COMPLETE',
        'UPDATE_COMPLETE',
        'UPDATE_ROLLBACK_FAILED',
        'UPDATE_ROLLBACK_COMPLETE'
    ]

    written_events = []

    while not complete:
        stack = get_stack_by_name(stack_name)
        if not stack:
            _print_event_log_separator()
            break

        if not written_events:
            _print_event_log_title()

        for event in reversed(CONNECTION.describe_stack_events(stack.stack_id)):
            # Don't print written events
            if event.event_id in written_events:
                continue

            # Don't print old events
            if event.timestamp < start_time:
                continue

            written_events.append(event.event_id)

            event_type, _ = event.resource_status.split('_', 1)
            log = False
            if not filter_type:
                log = True
            elif (filter_type == 'CREATE'
                    and event_type in ['CREATE', 'ROLLBACK']):
                log = True
            elif (filter_type == 'DELETE'
                    and event_type in ['DELETE', 'ROLLBACK']):
                log = True
            elif (filter_type == 'UPDATE'
                    and event_type in ['UPDATE', 'ROLLBACK']):
                log = True

            if not log:
                continue

            _print_event_log_event(event)

        if stack.stack_status in complete_statuses:
            _print_event_log_separator()
            LOGGER.info('Stack {} - Stack completed with status {}'.format(
                stack.stack_name,
                stack.stack_status))
            complete = True

        time.sleep(check_interval)



""" Bundling functions """
import fnmatch
import hashlib
import logging
import os
import subprocess
import sys
import tempfile
import zipfile

if sys.platform in ['win32', 'cygwin']:
    import ntpath as ospath
else:
    import os.path as ospath

from cumulus_ds import connection_handler
from cumulus_ds.config import CONFIG as config
from cumulus_ds.exceptions import (
    ChecksumMismatchException,
    HookExecutionException,
    UnsupportedCompression)

logger = logging.getLogger(__name__)


def build_bundles():
    """ Build bundles for the environment """
    bundle_types = config.get_bundles()

    if not bundle_types:
        logger.warning(
            'No bundles configured, will deploy without any bundles')
        return None

    for bundle_type in bundle_types:
        # Run pre-bundle-hook
        _pre_bundle_hook(bundle_type)

        if config.has_pre_built_bundle(bundle_type):
            bundle_path = config.get_pre_built_bundle_path(
                bundle_type)
            logger.info('Using pre-built bundle: {}'.format(bundle_path))

            try:
                _upload_bundle(bundle_path, bundle_type)
            except UnsupportedCompression:
                raise
        else:
            logger.info('Building bundle {}'.format(bundle_type))
            logger.debug('Bundle paths: {}'.format(', '.join(
                config.get_bundle_paths(bundle_type))))

            tmptar = tempfile.NamedTemporaryFile(
                suffix='.zip',
                delete=False)
            logger.debug('Created temporary tar file {}'.format(tmptar.name))

            try:
                _bundle_zip(
                    tmptar,
                    bundle_type,
                    config.get_environment(),
                    config.get_bundle_paths(bundle_type))

                tmptar.close()

                try:
                    _upload_bundle(tmptar.name, bundle_type)
                except UnsupportedCompression:
                    raise
            finally:
                logger.debug('Removing temporary tar file {}'.format(
                    tmptar.name))
                os.remove(tmptar.name)

        # Run post-bundle-hook
        _post_bundle_hook(bundle_type)


def _bundle_zip(tmpfile, bundle_type, environment, paths):
    """ Create a zip archive

    :type tmpfile: tempfile instance
    :param tmpfile: Tempfile object
    :type bundle_type: str
    :param bundle_type: Bundle name
    :type environment: str
    :param environment: Environment name
    :type paths: list
    :param paths: List of paths to include
    """
    archive = zipfile.ZipFile(tmpfile, 'w')
    path_rewrites = config.get_bundle_path_rewrites(bundle_type)

    for path in paths:
        for filename in _find_files(path, '*.*'):
            arcname = filename

            # Exclude files with other target environments
            prefix = '__cumulus-{}__'.format(environment)
            basename = ospath.basename(filename)

            if basename.startswith('__cumulus-'):
                if len(basename.split(prefix)) != 2:
                    logger.debug('Excluding file {}'.format(filename))
                    continue
            elif prefix in filename.split(ospath.sep):
                logger.debug('Excluding file {}'.format(filename))
                continue

            # Do all rewrites
            for rewrite in path_rewrites:
                target = rewrite['target'].replace('\\\\', '\\')
                destination = rewrite['destination'].replace('\\\\', '\\')

                try:
                    if arcname[:len(target)] == target:
                        arcname = arcname.replace(
                            target,
                            destination)
                        logger.debug(
                            'Replaced "{}" with "{}" in bundle {}'.format(
                                target,
                                destination,
                                bundle_type))
                except IndexError:
                    pass

            archive.write(filename, arcname, zipfile.ZIP_DEFLATED)

    archive.close()


def _find_files(directory, pattern):
    for root, dirs, files in os.walk(directory):
        for basename in files:
            if fnmatch.fnmatch(basename, pattern):
                filename = ospath.join(root, basename)
                yield filename


def _generate_local_md5hash(filename):
    """ Get the MD5 hash of a local file

    :type filename: str
    :param filename: Path to the file to read
    :returns: str -- MD5 of the file
    """
    if not ospath.exists(filename):
        logger.warning(
            'Unable to generate MD5 of local file {}. '
            'File does not exist.'.format(filename))
        return None

    hash = hashlib.md5(open(filename).read()).hexdigest()
    logger.debug('Generated md5 checksum for {} ({})'.format(
        ospath.basename(filename), hash))

    return hash


def _key_exists(bucket_name, key_name, checksum=None):
    """ Check if the given key exists in AWS S3.

    If checksum is given, also check if the md5 checksum is the same

    :type bucket_name: str
    :param bucket_name: S3 bucket name
    :type key_name: str
    :param key_name: S3 key name
    :type checksum: str
    :param checksum: MD5 checksum
    :returns: bool -- True if the key exists
    """
    try:
        connection = connection_handler.connect_s3()
    except Exception:
        raise

    bucket = connection.get_bucket(bucket_name)

    key = bucket.get_key(key_name)
    if not key:
        return False

    if checksum:
        if key.etag.replace('"', '') == checksum:
            return True
        return False

    return True


def _post_bundle_hook(bundle_name):
    """ Execute a post-bundle-hook

    :type bundle: str
    :param bundle: Bundle name
    """
    command = config.get_post_bundle_hook(bundle_name)

    if not command:
        return None

    logger.info('Running post-bundle-hook command: "{}"'.format(command))
    try:
        subprocess.check_call(command, shell=True)
    except subprocess.CalledProcessError, error:
        raise HookExecutionException(
            'The post-bundle-hook returned a non-zero exit code: {}'.format(
                error))


def _pre_bundle_hook(bundle_name):
    """ Execute a pre-bundle-hook

    :type bundle: str
    :param bundle: Bundle name
    """
    command = config.get_pre_bundle_hook(bundle_name)

    if not command:
        return None

    logger.info('Running pre-bundle-hook command: "{}"'.format(command))
    try:
        subprocess.check_call(command, shell=True)
    except subprocess.CalledProcessError, error:
        raise HookExecutionException(
            'The pre-bundle-hook returned a non-zero exit code: {}'.format(
                error))


def _upload_bundle(bundle_path, bundle_type):
    """ Upload all bundles to S3

    :type bundle_path: str
    :param bundle_path: Local path to the bundle
    :type bundle_type: str
    :param bundle_type: Bundle type
    """
    try:
        connection = connection_handler.connect_s3()
    except Exception:
        raise

    bucket = connection.get_bucket(
        config.get_environment_option('bucket'))

    # Check that the bundle actually exists
    if not ospath.exists(bundle_path):
        logger.error('File not found: {}'.format(bundle_path))
        sys.exit(1)

    if bundle_path.endswith('.zip'):
        compression = 'zip'
    else:
        raise UnsupportedCompression(
            'Unknown compression format for {}. '
            'We are currently only supporting .zip'.format(bundle_path))

    # Generate a md5 checksum for the local bundle
    local_hash = _generate_local_md5hash(bundle_path)

    key_name = (
        '{environment}/{version}/'
        'bundle-{environment}-{version}-{bundle_type}.{compression}').format(
            environment=config.get_environment(),
            version=config.get_environment_option('version'),
            bundle_type=bundle_type,
            compression=compression)

    # Do not upload bundles if the key already exists and has the same
    # md5 checksum
    if _key_exists(
            config.get_environment_option('bucket'),
            key_name,
            checksum=local_hash):
        logger.info(
            'This bundle is already uploaded to AWS S3. Skipping upload.')
        return

    # Get the key object
    key = bucket.new_key(key_name)

    logger.info('Starting upload of {} to s3://{}/{}'.format(
        ospath.basename(bundle_path), bucket.name, key_name))

    key.set_contents_from_filename(bundle_path, replace=True)

    logger.info('Completed upload of {} to s3://{}/{}'.format(
        ospath.basename(bundle_path), bucket.name, key_name))

    # Compare MD5 checksums
    if local_hash == key.md5:
        logger.debug('Uploaded bundle checksum OK ({})'.format(key.md5))
    else:
        logger.error('Mismatching md5 checksum {} ({}) and {} ({})'.format(
            bundle_path, local_hash, key_name, key.md5))
        raise ChecksumMismatchException(
            'Mismatching md5 checksum {} ({}) and {} ({})'.format(
                bundle_path, local_hash, key_name, key.md5))



""" Configuration management """
import logging
import sys
from ConfigParser import SafeConfigParser

if sys.platform in ['win32', 'cygwin']:
    import ntpath as ospath
else:
    import os.path as ospath

from cumulus_ds.config import config_file
from cumulus_ds.config.command_line_options import PARSER as parser
from cumulus_ds.exceptions import ConfigurationException

LOGGER = logging.getLogger(__name__)


class Configuration:
    environment = None
    config = None
    args = None

    def __init__(self):
        """ Constructor """
        self._parse_command_line_options()
        self._parse_configuration_file()
        self.environment = self.args.environment

    def _parse_command_line_options(self):
        """ Parse the command line options and populate self.args """
        self.args = parser.parse_args()

        # Make the stacks prettier
        if self.args.stacks:
            self.args.stacks = [s.strip() for s in self.args.stacks.split(',')]

        if self.args.cumulus_version:
            settings_conf = SafeConfigParser()
            settings_conf.read(
                ospath.realpath(
                    '{}/../settings.conf'.format(ospath.dirname(__file__))))
            print('Cumulus version {}'.format(
                settings_conf.get('general', 'version')))
            sys.exit(0)
        elif not self.args.environment:
            raise ConfigurationException('--environment is required')

    def _parse_configuration_file(self):
        """ Parse the configuration file """
        self.config = config_file.configure(self.args)

    def get_environment(self):
        """ Returns the environment name

        :returns: str
        """
        return unicode(self.environment)

    def get_bundle_path_rewrites(self, bundle):
        """ Returns a dict with all path rewrites

        :type bundle: str
        :param bundle: Bundle name
        :returns: dict
        """
        try:
            return self.config['bundles'][bundle]['path-rewrites']
        except KeyError:
            return []

    def get_bundle_paths(self, bundle):
        """ Returns a list of bundle paths for a given bundle

        :type bundle: str
        :param bundle: Bundle name
        :returns: list
        """
        try:
            return self.config['bundles'][bundle]['paths']
        except KeyError:
            LOGGER.error('No paths defined for bundle "{}"'.format(bundle))
            raise ConfigurationException(
                'No paths defined for bundle "{}"'.format(bundle))

    def get_bundles(self):
        """ Returns a list of bundles"""
        try:
            bundles = self.config['environments'][self.environment]['bundles']
        except KeyError:
            LOGGER.warning(
                'No bundles found for environment {}'.format(self.environment))
            return None

        # Check that the bundles are configured
        for bundle in bundles:
            if not bundle:
                bundles.remove(bundle)
                continue

            if bundle not in self.config['bundles']:
                bundles.remove(bundle)
                LOGGER.warning(
                    'No matching configuration for bundle "{}"!'.format(bundle))

        return bundles

    def get_environment_option(self, option_name):
        """ Returns version number

        :returns: str
        """
        try:
            return self.config['environments'][self.environment][option_name]
        except KeyError:
            LOGGER.error('No option {} in environment {}'.format(
                option_name, self.environment))
            return None

    def get_log_level(self):
        """ Returns the log level

        :returns: str
        """
        try:
            return self.config['general']['log-level']
        except KeyError:
            return 'DEBUG'

    def get_post_bundle_hook(self, bundle):
        """ Returns the post bundle hook command or None

        :type bundle: str
        :param bundle: Bundle name
        :returns: str or None
        """
        try:
            return self.config['bundles'][bundle]['post-bundle-hook']
        except KeyError:
            return None

    def get_post_deploy_hook(self):
        """ Returns the post deploy hook command or None

        :returns: str or None
        """
        try:
            return self.config[
                'environments'][self.environment]['post-deploy-hook']
        except KeyError:
            return None

    def get_pre_built_bundle_path(self, bundle):
        """ Return the path to the pre-built bundle

        :type bundle: str
        :param bundle: Bundle name
        :returns: str -- Path to the pre-built bundle
        """
        try:
            return str(self.config['bundles'][bundle]['pre-built-bundle'])
        except KeyError:
            LOGGER.error('No pre-built bundle path found for "{}"'.format(
                bundle))
            raise ConfigurationException(
                'No pre-built bundle path found for "{}"'.format(bundle))

    def get_pre_deploy_hook(self):
        """ Returns the pre deploy hook command or None

        :returns: str or None
        """
        try:
            return self.config[
                'environments'][self.environment]['pre-deploy-hook']
        except KeyError:
            return None

    def get_pre_bundle_hook(self, bundle):
        """ Returns the pre bundle hook command or None

        :type bundle: str
        :param bundle: Bundle name
        :returns: str or None
        """
        try:
            return self.config['bundles'][bundle]['pre-bundle-hook']
        except KeyError:
            return None

    def get_stack_disable_rollback(self, stack):
        """ See if we should disable rollback

        :type stack: str
        :param stack: Stack name
        :returns: bool -- Disable rollback?
        """
        try:
            return self.config['stacks'][stack]['disable-rollback']
        except KeyError:
            raise ConfigurationException(
                'Stack template not found in configuration')

    def get_stack_parameters(self, stack):
        """ Return the stack parameters

        :type stack: str
        :param stack: Stack name
        :returns: list -- All stack parameters
        """
        try:
            return self.config['stacks'][stack]['parameters']
        except KeyError:
            return []

    def get_stack_tags(self, stack):
        """ Return the stack tags

        :type stack: str
        :param stack: Stack name
        :returns: list -- All stack tags
        """
        try:
            return self.config['stacks'][stack]['tags']
        except KeyError:
            return None

    def get_stack_template(self, stack):
        """ Return the path to the stack template

        :type stack: str
        :param stack: Stack name
        :returns: str -- Stack template path
        """
        try:
            return self.config['stacks'][stack]['template']
        except KeyError:
            raise ConfigurationException(
                'Stack template not found in configuration for stack {}'.format(
                    stack))

    def get_stack_timeout_in_minutes(self, stack):
        """ Return stack creation timeout

        :type stack: str
        :param stack: Stack name
        :returns: int -- Stack creation timeout in minutes
        """
        try:
            timeout = int(self.config['stacks'][stack]['timeout-in-minutes'])
        except KeyError:
            raise ConfigurationException(
                'Stack timeout not found in configuration')

        if timeout == 0:
            return None
        return timeout

    def get_stacks(self):
        """ Returns a list of stacks """
        try:
            return self.config['stacks'].keys()
        except KeyError:
            LOGGER.warning(
                'No stacks found for environment {}'.format(self.environment))
            return None

    def has_pre_built_bundle(self, bundle):
        """ Checks wether or not the bundle has a pre-built-bundle flag

        :type bundle: str
        :param bundle: Bundle name
        :returns: bool -- True if there is a pre-built-bundle
        """
        if 'pre-built-bundle' in self.config['bundles'][bundle]:
            return True
        return False


CONFIG = Configuration()



""" The deployment manager handles all deployments of CloudFormation stacks """
import logging
import subprocess

from cumulus_ds import terminal_size
from cumulus_ds.config import CONFIG as config
from cumulus_ds.helpers.stack import (
    delete_stack,
    ensure_stack,
    list_events_all_stacks,
    list_all_stacks,
    print_output_all_stacks,
    validate_templates_all_stacks)
from cumulus_ds.exceptions import HookExecutionException

LOGGER = logging.getLogger(__name__)

TERMINAL_WIDTH, _ = terminal_size.get_terminal_size()


def deploy():
    """ Ensure stack is up and running (create or update it) """
    # Run pre-deploy-hook
    _pre_deploy_hook()

    stack_names = config.get_stacks()

    if not stack_names:
        LOGGER.warning('No stacks configured, nothing to deploy')
        return

    for stack_name in stack_names:
        ensure_stack(
            stack_name,
            template=config.get_stack_template(stack_name),
            disable_rollback=config.get_stack_disable_rollback(stack_name),
            parameters=config.get_stack_parameters(stack_name),
            timeout_in_minutes=config.get_stack_timeout_in_minutes(stack_name),
            tags=config.get_stack_tags(stack_name))

    # Run post-deploy-hook
    _post_deploy_hook()


def list_events():
    """ List events """
    list_events_all_stacks()


def list_outputs():
    """ List all outputs for all stacks """
    print_output_all_stacks()


def list_stacks():
    """ List stacks and their statuses """
    list_all_stacks()


def undeploy(force=False):
    """ Undeploy an environment

    :type force: bool
    :param force: Skip the safety question
    """
    message = (
        'This will DELETE all stacks in the environment. '
        'This action cannot be undone. '
        'Are you sure you want to do continue? [N/y] ')

    choice = 'yes'
    if not force:
        choice = raw_input(message).lower()
        if choice not in ['yes', 'y']:
            print('Skipping undeployment.')
            return None

    stacks = config.get_stacks()
    stacks.reverse()

    if not stacks:
        LOGGER.warning('No stacks to undeploy.')
        return None

    for stack in stacks:
        delete_stack(stack)


def validate_templates():
    """ Validate the template """
    validate_templates_all_stacks()


def _pre_deploy_hook():
    """ Execute a pre-deploy-hook """
    command = config.get_pre_deploy_hook()

    if not command:
        return None

    LOGGER.info('Running pre-deploy-hook command: "{}"'.format(command))
    try:
        subprocess.check_call(command, shell=True)
    except subprocess.CalledProcessError, error:
        raise HookExecutionException(
            'The pre-deploy-hook returned a non-zero exit code: {}'.format(
                error))


def _post_deploy_hook():
    """ Execute a post-deploy-hook """
    command = config.get_post_deploy_hook()

    if not command:
        return None

    LOGGER.info('Running post-deploy-hook command: "{}"'.format(command))
    try:
        subprocess.check_call(command, shell=True)
    except subprocess.CalledProcessError, error:
        raise HookExecutionException(
            'The post-deploy-hook returned a non-zero exit code: {}'.format(
                error))



""" Setup script for PyPI """
import os
from setuptools import setup
from ConfigParser import SafeConfigParser

settings = SafeConfigParser()
settings.read(os.path.realpath('cumulus_ds/settings.conf'))


setup(
    name='cumulus',
    version='{}b'.format(settings.get('general', 'version')),
    license='Apache Software License',
    description='Cumulus Deployment Suite for Amazon Web Services',
    author='Sebastian Dahlgren',
    author_email='sebastian.dahlgren@skymill.se',
    url='http://www.skymillsolutions.com',
    keywords="cumulus cloudformation amazon web services",
    platforms=['Any'],
    packages=['cumulus_ds'],
    scripts=['cumulus'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'boto >= 2.12.0'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python'
    ]
)



""" Manage configuration file parsing """
import logging
import os
import sys
from ConfigParser import SafeConfigParser, NoOptionError

if sys.platform in ['win32', 'cygwin']:
    import ntpath as ospath
else:
    import os.path as ospath

from cumulus_ds.exceptions import ConfigurationException

LOGGER = logging.getLogger(__name__)

# Initial configuration object
CONF = {
    'general': {},
    'environments': {},
    'stacks': {},
    'bundles': {}
}

# Options per section
# [(option, required)]
GENERAL_OPTIONS = [
    ('log-level', False)
]
STACK_OPTIONS = [
    ('template', True),
    ('disable-rollback', True),
    ('parameters', False),
    ('timeout-in-minutes', False),
    ('tags', False)
]
BUNDLE_OPTIONS = [
    ('paths', True),
    ('path-rewrites', False),
    ('pre-bundle-hook', False),
    ('post-bundle-hook', False),
    ('pre-built-bundle', False)
]
ENV_OPTIONS = [
    ('access-key-id', True),
    ('secret-access-key', True),
    ('bucket', True),
    ('region', True),
    ('stacks', True),
    ('bundles', True),
    ('version', False),
    ('pre-deploy-hook', False),
    ('post-deploy-hook', False),
    ('stack-name-prefix', False),
    ('stack-name-suffix', False)
]


def configure(args):
    """ Populate the objects

    :type args: Namespace
    :param args: Parsed arguments from argparse
    :type config_file: str or None
    :param config_file: Configuration file to read from
    """
    config_files = []

    if sys.platform in ['win32', 'cygwin']:
        config_files.append('C:\\cumulus.conf')
        config_files.append(ospath.expanduser('~\\.cumulus.conf'))
        config_files.append('{}\\cumulus.conf'.format(os.curdir))
    else:
        config_files.append('/etc/cumulus.conf')
        config_files.append(ospath.expanduser('~/.cumulus.conf'))
        config_files.append('{}/cumulus.conf'.format(os.curdir))

    # Add custom configuration file path
    if args.config:
        if ospath.exists(ospath.expanduser(args.config)):
            config_files = [ospath.expanduser(args.config)]
        else:
            LOGGER.warning('Configuration file {} not found.'.format(
                ospath.expanduser(args.config)))

    # Read config file
    conf_file_found = False
    for conf_file in config_files:
        if ospath.exists(conf_file):
            conf_file_found = True
            LOGGER.info('Reading configuration from {}'.format(conf_file))
    if not conf_file_found:
        raise ConfigurationException(
            'No configuration file found. Looked for {}'.format(
                ', '.join(config_files)))

    config = SafeConfigParser()
    config.read(config_files)

    _populate_general(args, config)
    _populate_environments(args, config)
    _populate_stacks(args, config)
    _populate_bundles(args, config)

    return CONF


def _populate_environments(args, config):
    """ Populate the environments config object

    :type args: Namespace
    :param args: Parsed arguments from argparse
    :type config: ConfigParser.read
    :param config: Config parser config object
    """
    section = 'environment: {}'.format(args.environment)
    environment = args.environment
    if not section in config.sections():
        raise ConfigurationException(
            'No configuration found for environment {}'.format(environment))

    CONF['environments'][environment] = {}

    for option, required in ENV_OPTIONS:
        try:
            if option == 'bundles':
                bundles = []
                for item in config.get(section, option).split(','):
                    bundles.append(item.strip())
                CONF['environments'][environment][option] = bundles
            elif option == 'stacks':
                stacks = []
                for item in config.get(section, option).split(','):
                    item = item.strip()

                    # If --stacks has been used,
                    # do only add those stacks
                    if args.stacks and item not in args.stacks:
                        continue

                    stacks.append('{}-{}'.format(
                        args.environment, item))
                CONF['environments'][environment][option] = stacks
            elif option == 'version':
                if args.version:
                    CONF['environments'][environment][option] = args.version
                else:
                    CONF['environments'][environment][option] = config.get(
                        section, option)
            else:
                CONF['environments'][environment][option] = \
                    config.get(section, option)
        except NoOptionError:
            if required:
                raise ConfigurationException(
                    'Missing required option {}'.format(option))


def _populate_general(args, config):
    """ Populate the general config object

    :type args: Namespace
    :param args: Parsed arguments from argparse
    :type config: ConfigParser.read
    :param config: Config parser config object
    """
    section = 'general'
    if not section in config.sections():
        return None

    for option, required in GENERAL_OPTIONS:
        try:
            if option == 'log-level':
                log_level = config.get(section, option).upper()

                if log_level not in ['DEBUG', 'INFO', 'WARNING', 'ERROR']:
                    LOGGER.warning(
                        (
                            'Invalid log level "{}". Using default log level.'
                        ).format(log_level))
                    log_level = 'DEBUG'

                CONF['general'][option] = log_level
            else:
                CONF['general'][option] = config.get(section, option)
        except NoOptionError:
            if required:
                raise ConfigurationException(
                    'Missing required option {}'.format(option))


def _populate_stacks(args, config):
    """ Populate the stacks config object

    :type args: Namespace
    :param args: Parsed arguments from argparse
    :type config: ConfigParser.read
    :param config: Config parser config object
    """
    for section in config.sections():
        if section.startswith('stack: '):
            stack = section.split(': ')[1]

            # If --stacks has been used, do only add those stacks
            if args.stacks and stack not in args.stacks:
                continue

            stack = '{}-{}'.format(args.environment, section.split(': ')[1])

            # Prepend a stack name prefix
            if 'stack-name-prefix' in CONF['environments'][args.environment]:
                stack = '{}-{}'.format(
                    CONF['environments'][args.environment]['stack-name-prefix'],
                    stack)

            # Append a stack name suffix
            if 'stack-name-suffix' in CONF['environments'][args.environment]:
                stack = '{}-{}'.format(
                    stack,
                    CONF['environments'][args.environment]['stack-name-suffix'])

            CONF['stacks'][stack] = {}

            for option, required in STACK_OPTIONS:
                try:
                    if option == 'disable-rollback':
                        CONF['stacks'][stack][option] = config.getboolean(
                            section, option)
                    elif option == 'template':
                        CONF['stacks'][stack][option] = ospath.expanduser(
                            config.get(section, option))
                    elif option == 'parameters':
                        try:
                            raw_parameters = config.get(section, option)\
                                .split('\n')
                            if not raw_parameters[0]:
                                raw_parameters.pop(0)

                            parameters = []
                            for parameter in raw_parameters:
                                key, value = parameter.split('=')
                                parameters.append((key.strip(), value.strip()))
                            CONF['stacks'][stack][option] = parameters
                        except ValueError:
                            raise ConfigurationException(
                                'Error parsing parameters for stack {}'.format(
                                    stack))
                    elif option == 'tags':
                        try:
                            raw_tags = config.get(section, option)\
                                .split('\n')
                            if not raw_tags[0]:
                                raw_tags.pop(0)

                            tags = {}
                            for tag in raw_tags:
                                key, value = tag.split('=')
                                tags[key.strip()] = value.strip()
                            CONF['stacks'][stack][option] = tags
                        except ValueError:
                            raise ConfigurationException(
                                'Error parsing tags for stack {}'.format(
                                    stack))
                    elif option == 'timeout-in-minutes':
                        CONF['stacks'][stack][option] = config.getint(
                            section, option)
                    else:
                        CONF['stacks'][stack][option] = config.get(
                            section, option)
                except NoOptionError:
                    if required:
                        raise ConfigurationException(
                            'Missing required option {}'.format(
                                option))

                    if option == 'timeout-in-minutes':
                        CONF['stacks'][stack][option] = 0

            # Add command line parameters
            try:
                if args.parameters:
                    if not 'parameters' in CONF['stacks'][stack]:
                        CONF['stacks'][stack]['parameters'] = []

                    for raw_parameter in args.parameters.split(','):
                        stack_name, keyvalue = raw_parameter.split(':')
                        key, value = keyvalue.split('=')
                        if stack_name == stack:
                            CONF['stacks'][stack]['parameters'].append(
                                (key, value))
            except ValueError:
                raise ConfigurationException('Error parsing --parameters')


def _populate_bundles(args, config):
    """ Populate the bundles config object

    :type args: Namespace
    :param args: Parsed arguments from argparse
    :type config: ConfigParser.read
    :param config: Config parser config object
    """
    for section in config.sections():
        if section.startswith('bundle: '):
            bundle = section.split(': ')[1]
            CONF['bundles'][bundle] = {}

            for option, required in BUNDLE_OPTIONS:
                try:
                    if option == 'paths':
                        lines = config.get(section, option).strip().split('\n')
                        paths = []
                        for path in lines:
                            paths.append(ospath.expanduser(path.strip()))
                        CONF['bundles'][bundle]['paths'] = paths
                    elif option == 'path-rewrites':
                        CONF['bundles'][bundle]['path-rewrites'] = []
                        lines = config.get(section, option).strip().split('\n')

                        for line in lines:
                            try:
                                target, destination = line.split('->')
                            except ValueError:
                                raise ConfigurationException(
                                    'Invalid path-rewrites for '
                                    'bundle {}'.format(bundle))

                            # Clean the target and destination from initial /
                            if target[0] == '/':
                                target = target[1:]
                            if destination[0] == '/':
                                destination = destination[1:]

                            CONF['bundles'][bundle]['path-rewrites'].append({
                                'target': target.strip(),
                                'destination': destination.strip()
                            })

                    else:
                        CONF['bundles'][bundle][option] = config.get(
                            section, option)
                except NoOptionError:
                    if (option == 'paths' and
                            config.has_option(section, 'pre-built-bundle')):
                        continue

                    if required:
                        raise ConfigurationException(
                            'Missing required option {}'.format(option))



""" Cumulus Deployment Suite

APACHE LICENSE 2.0
Copyright 2013-2014 Skymill Solutions

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging

from cumulus_ds import bundle_manager
from cumulus_ds import deployment_manager
from cumulus_ds.config import CONFIG as config

LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format':
            '%(asctime)s - cumulus - %(levelname)s - %(message)s'
        },
        'boto': {
            'format':
            '%(asctime)s - boto - %(levelname)s - %(message)s'
        }
    },
    'handlers': {
        'default': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'standard'
        },
        'boto': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'boto'
        }
    },
    'loggers': {
        '': {
            'handlers': ['default'],
            'level': 'DEBUG',
            'propagate': True
        },
        'boto': {
            'handlers': ['boto'],
            'level': logging.CRITICAL,
            'propagate': False
        },
        'cumulus_ds.bundle_manager': {
            'handlers': ['default'],
            'level': 'DEBUG',
            'propagate': False
        },
        'cumulus_ds.config_handler': {
            'handlers': ['default'],
            'level': 'DEBUG',
            'propagate': False
        },
        'cumulus_ds.connection_handler': {
            'handlers': ['default'],
            'level': 'DEBUG',
            'propagate': False
        },
        'cumulus_ds.deployment_manager': {
            'handlers': ['default'],
            'level': 'DEBUG',
            'propagate': False
        },
        'cumulus_ds.helpers.stack': {
            'handlers': ['default'],
            'level': 'DEBUG',
            'propagate': False
        }
    }
}

# Set log level
LOGGING_CONFIG['handlers']['default']['level'] = config.get_log_level()

logging.config.dictConfig(LOGGING_CONFIG)
LOGGER = logging.getLogger(__name__)


def main():
    """ Main function """
    try:
        if config.args.bundle:
            bundle_manager.build_bundles()

        if config.args.deploy:
            bundle_manager.build_bundles()
            deployment_manager.deploy()

        if config.args.deploy_without_bundling:
            deployment_manager.deploy()

        if config.args.list:
            deployment_manager.list_stacks()

        if config.args.undeploy:
            deployment_manager.undeploy(force=config.args.force)

        if config.args.validate_templates:
            deployment_manager.validate_templates()

        if config.args.events:
            deployment_manager.list_events()

        if config.args.outputs:
            deployment_manager.list_outputs()

    except Exception as error:
        LOGGER.error(error)
        raise
